/*****
 *
 * This file is part of libcss.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****/

/*
 * css_dvd: Software emulation of Drive and Host(Player) CSS
 */

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>

#include "css_dvd.h"
#include "css_auth.h"
#include "css.h"

typedef struct  {
	u_int32_t lba;
	u_int8_t key[LEN_KEY];
} titlekey_t;

struct cssparams {
	int agid_busy[4];	// only used by drive emulation: wich agid is used?
	int keys_present;	// is there a valid disc key?
	u_int8_t buskey[LEN_KEY];	// intermediate key
	u_int8_t disckey[2048];	// the disc key, unobfuscated

	titlekey_t titlekey_o[200];// all the title keys, obfuscated only by disc key
	titlekey_t titlekey[200];// all the title keys, unobfuscated
	int titlekeys;		// number of title keys in titlekey[]

	u_int8_t Key1[LEN_KEY];	// for authorisation
	u_int8_t Key2[LEN_KEY];	// for authorisation
	u_int8_t KeyCheck[LEN_KEY];	// for authorisation

	u_int8_t challenge[10];	// for authorisation
	u_int8_t response[5];	// for authorisation

	int authOK;		// first of 2 steps of auth already done?
	int varient;		// for authorisation, choosed by the drive, guessed by the player
} CSSDrive, CSSHost;	// 2 sets of parameters, one each for drive and host


/**
 * Load and save key set from/to disc
 *
 * reads disc and title keys from a file with the extension .CSS
 * returns 0 on success, -1 on error
 **/

int CSSDriveReadKeys (char *_filename) 
{
	char filename[PATH_MAX];
	int fd;
	int i;
	char buf[3];

	strncpy (filename, _filename, PATH_MAX);
	filename[PATH_MAX] = '\0';	// just in case

	// does filename end on 000?
	if (!strncmp(&filename[strlen(filename) - 4], ".000", 4))
		filename[strlen(filename) - 4] = '\0';	// then cut off extension

	// does filename end on CSS?
	if (strncasecmp (&filename[strlen(filename) - 4], ".css", 4))
		strcat(filename, ".css");	// if not, append it

	if ((fd = open (filename, O_RDONLY)) < 0)
		return -1;

	// check file identifier
	read (fd, buf, 3); 
	if (strncmp (buf, "CSS", 3))
		return -1;

	// read disc key
	read (fd, CSSDrive.disckey, 2048);

	// read title keys
	for (i=0; read (fd, &CSSDrive.titlekey_o[i].lba, sizeof CSSDrive.titlekey_o[i].lba); i++) {
		read (fd, CSSDrive.titlekey_o[i].key, LEN_KEY);
		CSSDrive.titlekey[i].lba = CSSDrive.titlekey_o[i].lba;
		memcpy(CSSDrive.titlekey[i].key, CSSDrive.titlekey_o[i].key, LEN_KEY);
		CSSDecryptTitleKey(CSSDrive.titlekey[i].key, CSSDrive.disckey);
	}

	CSSDrive.titlekeys = i;
	CSSDrive.keys_present = 1;

	close (fd);

	return 0;
}


/**
 *
 **/

int CSSHostWriteKeys (char *_filename) 
{
	char filename[PATH_MAX];
	int fd;
	int i;

	strncpy (filename, _filename, PATH_MAX);
	filename[PATH_MAX] = '\0';	// just in case

	// does Filename end on 000?
	if (!strncmp (&filename[strlen(filename) - 4], ".000", 4))
		filename[strlen(filename) - 4] = '\0';	// then cut off extension

	// does Filename end on CSS?
	if (strncasecmp (&filename[strlen(filename) - 4], ".css", 4))
		strcat (filename, ".css");	// if not, append it

	if ((fd = open (filename, O_CREAT | O_WRONLY)) < 0)
		return 1;

	// Write ID to file
	write (fd, "CSS", 3);

	// Write disc key to file
	write (fd, CSSHost.disckey, 2048);

	// Write title keys to file
	for (i=0; i<CSSHost.titlekeys; i++) {
		write (fd, &CSSHost.titlekey_o[i].lba, sizeof CSSHost.titlekey_o[i].lba);
		write (fd, CSSHost.titlekey_o[i].key, LEN_KEY);
	}

	close (fd);

	return 0;
}


/**
 * drive authorisation: AGID
 *
 * Invalidates previously released AGID
 * returns 0 on success, 1 on error
 **/

int CSSDriveInvalidateAGID (int agid) 
{
	if ((agid & 3) != agid)		// agid out of range
		return 1;

	CSSDrive.agid_busy[agid] = 0;

	return 0;
}


/**
 * Initializes the CSS process with the drive
 * returns positive AGID on success, <0 on error
 **/

int CSSDriveRequestAGID (void) 
{
	int i;

	for (i=3; (i>=0) && CSSDrive.agid_busy[i]; i--);
	if ((i<0) || CSSDrive.agid_busy[i])
		return -1;

	CSSDrive.agid_busy[i] = 1;
	CSSDrive.varient = -1;
	CSSDrive.authOK = 0;

	return i;
}


/**
 * Reset auth process for host
 * returns 0 on success
 **/

int CSSHostInitAuth (void) 
{
	CSSHost.varient = -1;
	CSSHost.authOK = 0;

	return 0;
}


/**
 * drive authorisation: bus key calculation
 *
 * Crypt Key1 and Key2 with varient to bus key
 * returns 0 on success, 1 on error
 **/

int CSSHostBusKey (void) 
{
	u_int8_t data[10];

	if (!CSSHost.authOK)
		return 1;

	memcpy (data, CSSHost.Key1, LEN_KEY);
	memcpy (data + LEN_KEY, CSSHost.Key2, LEN_KEY);

	CryptBusKey (CSSHost.varient, data, CSSHost.KeyCheck);

	memcpy (CSSHost.buskey, CSSHost.KeyCheck, LEN_KEY);

	CSSHost.authOK = 0;

	return 0;
}


/**
 * Crypt Key1 and Key2 with varient to bus key
 * returns 0 on success, 1 on error
 **/

int CSSDriveBusKey (void) 
{
	u_int8_t data[10];

	if (!CSSDrive.authOK)
		return 1;

	memcpy (data, CSSDrive.Key1, LEN_KEY);
	memcpy (data + LEN_KEY, CSSDrive.Key2, LEN_KEY);

	CryptBusKey (CSSDrive.varient, data, CSSDrive.KeyCheck);

	memcpy (CSSDrive.buskey, CSSDrive.KeyCheck, LEN_KEY);

	CSSDrive.authOK = 0;

	return 0;
}


/**
 * drive authorisation: challenge & response
 *
 * Host part:
 * (attention: name inconsistency. CSSDrive used despite Host function)
 * retreive 10 byte challenge from host for drive
 * returns 0 on success, 1 on error
 **/

int CSSDriveAuthChallenge (u_int8_t *data) 
{
	int i;

	if (!data)
		return 1;

	// Make up 10 bytes of challenge data
	for (i = 0; i < 10; i++)
		data[i] = CSSHost.challenge[9 - i] = rand() & 0xFF;

	return 0;
}


/**
 * Drive part:
 * Post host challenge (10 bytes) into drive
 * and retreive drive response (5 bytes == LEN_KEY) for host
 * returns 0 on success, 1 on error
 **/

int CSSDriveAuth (u_int8_t *data) 
{
	int i;

	if (!data)
		return 1;

	if (!CSSDrive.keys_present)		// no key file
		return 1;

	for (i = 0; i < 10; i++)
		CSSDrive.challenge[9-i] = data[i];

	// Make up varient for this authorisation, if not already present
	if (CSSDrive.varient < 0)
		CSSDrive.varient = rand() & 0x1F;

	// Calculate drive response for challenge from host
	CryptKey1(CSSDrive.varient, CSSDrive.challenge, CSSDrive.Key1);

	for (i = 0; i < LEN_KEY; i++)
		data[i] = CSSDrive.Key1[4 - i];

	// Calculate bus key if we did other part of auth 
	if (CSSDrive.authOK)
		CSSDriveBusKey();

	else CSSDrive.authOK = 1;

	return 0;
}


/**
 * Host part:
 * (attention: name inconsistency. CSSDrive used despite Host function)
 * Verify 5 bytes response from drive
 * returns 0 on success, -1 on failure
 **/

int CSSDriveAuthResponse (u_int8_t *data) 
{
	int i;

	if (!data)
		return -1;

	for (i=0; i<LEN_KEY; i++)
		CSSHost.Key1[4-i] = data[i];

	// Guess 5 bit varient
	for (i=0; i<32; i++) {
		CryptKey1 (i, CSSHost.challenge, CSSHost.KeyCheck);

		if (!memcmp (CSSHost.KeyCheck, CSSHost.Key1, LEN_KEY))
			goto varient_found;
	}

	// If no varient matched, the response was wrong.
	return -1;

varient_found:
	CSSHost.varient = i;

	// Calculate bus key if we did other part of auth 
	if (CSSHost.authOK)
		CSSHostBusKey();
	else
		CSSHost.authOK = 1;

	return 0;
}


/**
 * host authorisation: challenge & response
 *
 * Drive part:
 * (attention: name inconsistency. CSSHost used despite Drive function)
 * Retreive drive challenge (10 bytes) 
 * returns 0 on success, -1 on error
 **/

int CSSHostAuthChallenge (u_int8_t *data) 
{
	int i;

	if (!data)
		return -1;

	if (!CSSDrive.keys_present)		// no key file
		return -2;

	// Make up 10 bytes of challenge data
	for (i=0; i<10; i++)
		data[i] = CSSDrive.challenge[9 - i] = rand() & 0x00FF;

	return 0;
}


/**
 * Host part:
 * Post drive challenge (10 bytes) into host
 * and retreive host response (5 bytes) for drive
 * returns 0 on success, 1 on error
 **/

int CSSHostAuth (u_int8_t *data) 
{
	int i;

	if (!data)
		return -1;

	for (i=0; i<10; i++)
		CSSHost.challenge[9 - i] = data[i];

	// Make up varient for this authorisation, if not already present
	if (CSSHost.varient < 0)
		CSSHost.varient = rand() & 0x1F;

	// Calculate host response for challenge from drive
	CryptKey2(CSSHost.varient, CSSHost.challenge, CSSHost.Key2);
	for (i=0; i<LEN_KEY; i++)
		data[i] = CSSHost.Key2[4-i];

	// Calculate bus key if we did other part of auth 
	if (CSSHost.authOK)
		CSSHostBusKey();
	else
		CSSHost.authOK = 1;

	return 0;
}


/**
 * Drive part:
 * (attention: name inconsistency. CSSHost used despite Drive function)
 * Post host response (5 bytes) into drive
 * returns 0 on success, -1 on error
 **/

int CSSHostAuthResponse (u_int8_t *data) 
{
	int i;

	if (!data)
		return -1;

	if (!CSSDrive.keys_present)		// no key file
		return -1;

	for (i=0; i<LEN_KEY; i++)
		CSSDrive.response[4-i] = data[i];

	// Guess 5 bit varient
	for (i=0; i<32; i++) {
		CryptKey2(i, CSSDrive.challenge, CSSDrive.Key2);
		if (memcmp(CSSDrive.Key2, CSSDrive.response, LEN_KEY) == 0) break;
	}

	// If no varient matched, the response was wrong.
	if (i>=32)
		return -1;

	CSSDrive.varient = i;

	// Calculate bus key if we did other part of auth 
	if (CSSDrive.authOK)
		CSSDriveBusKey();
	else
		CSSDrive.authOK = 1;

	return 0;
}


/**
 * disc key transfer
 *
 * Retreive bus encrypted disc key (2048 byte) from drive
 * returns 0 on success, 1 on error
 **/

int CSSDriveDiscKey (u_int8_t *data) 
{
	int i;

	if (!data)
		return -1;

	if (!CSSDrive.keys_present)		// no key file
		return -1;

	// obfuscate disc key with bus key
	for (i=0; i < 2048; i++)
		data[i] = CSSDrive.disckey[i] ^ CSSDrive.buskey[4 - (i % LEN_KEY)];

	return 0;
}


/**
 * post bus encrypted disc key (2048 byte) into host
 * returns 0
 **/

int CSSHostDiscKey (u_int8_t *data) 
{
	int i;

	// decrypt disc key
	for (i=0; i < 2048; i++)
		CSSHost.disckey[i] = data[i] ^ CSSHost.buskey[4 - (i % LEN_KEY)];

	CSSHost.keys_present = 1;

	return 0;
}


/**
 * title key transfer
 *
 * Retreive bus encrypted title key (5 byte) from drive
 * input parameters:
 *   lba: absolute number of logical block containing the title key
 * return parameters:
 *   data: will be filled with the bus obfuscated title key
 *   cpm: 0=material not copyrighted, 1=copyrighted material
 *   cp_sec: 1=this sector has specific structure for copyright protection system, 0=it doesn't
 *   cgms: 0=copying is permitted, 2=one generation of copies may be made, 3=no copying is allowed
 * returns 0 on success, 1 on error
 **/

int CSSDriveTitleKey (u_int32_t lba, u_int8_t *data, int *cpm, int *cp_sec, int *cgms) 
{
	int i;
	int keynum;

	if (!data)
		return -1;

	if (!CSSDrive.keys_present)		// no key file
		return -1;

	// find appropiate title key for this lba
	if (CSSDrive.titlekeys <= 0)		// no title keys
		return -1;

	keynum = CSSDrive.titlekeys - 1;
	while ((keynum >= 0) && (CSSDrive.titlekey_o[keynum].lba > lba))
		keynum--;

	// error, key ranges starts after this lba: try the first key instead
	if (keynum < 0)
		keynum = 0;
	
	// obfuscate with bus key
	for (i=0; i < LEN_KEY; i++)
		data[i] = CSSDrive.titlekey_o[keynum].key[i] ^ CSSDrive.buskey[4 - i];

	if (cpm)
		*cpm = 0;
	if (cp_sec)
		*cp_sec = 0;
	if (cgms)
		*cgms = 0;

	return 0;
}


/**
 * post bus encrypted title key (5 byte) into host
 * returns 0
 **/

int CSSHostTitleKey (u_int32_t lba, u_int8_t *data) 
{
	int i;

	// decrypt title key
	for (i = 0; i < LEN_KEY; i++)
		CSSHost.titlekey_o[CSSHost.titlekeys].key[i] = data[i] ^ CSSHost.buskey[4 - i];

	// first title key or different from previous one?
	if ((!CSSHost.titlekeys) || (memcmp(CSSHost.titlekey_o[CSSHost.titlekeys - 1].key, CSSHost.titlekey_o[CSSHost.titlekeys].key, LEN_KEY))) {
		// store lba and increment key count
		CSSHost.titlekey[CSSHost.titlekeys].lba = lba;
		memcpy(CSSHost.titlekey[CSSHost.titlekeys].key, CSSHost.titlekey_o[CSSHost.titlekeys].key, LEN_KEY);
		CSSDecryptTitleKey(CSSHost.titlekey[CSSHost.titlekeys].key, CSSHost.disckey);
		CSSHost.titlekey_o[CSSHost.titlekeys++].lba = lba;
	}

	return 0;
}


/**
 * content decoding / encoding
 *
 * Decrypts CSS encrypted blocks
 * lb_number: LBA of the first block
 * blockcount: number of blocks
 * buffer: pointer to the blocks
 * returns 0 if everything went ok
 *   -1 on error (failed to decrypt)
 **/

int CSSHostDecrypt (u_int32_t lb_number, u_int blockcount, u_int8_t *buffer) 
{
	int i, keynum, padding;
	u_int8_t *lb;
	u_int32_t lba;
	static u_int8_t packstartcode[] = {0x00, 0x00, 0x01, 0xBA};

	for (i = 0; i < blockcount; i++) {
		lba = lb_number + i;
		lb = &buffer[i * 2048];

		// check for pack start code at block start
		if (memcmp(lb, packstartcode, 4)) 
			continue;
		padding = lb[13] & 0x07;

		// check for start code (0x000001) at packet start
		if (memcmp(&lb[padding + 14], packstartcode, 3)) 
			continue;

		// PES packet?
		if (lb[padding + 17] < 0xBB) 
			continue;

		// PES encrypted?
		if ((lb[padding + 20] & 0x30) != 0x10) 
			continue;

		// which title key to use?
		keynum = CSSHost.titlekeys - 1;
		while ((keynum >= 0) && (CSSHost.titlekey[keynum].lba > lba))
			keynum--;

		// error, key ranges start after this lba: try the first key instead
		if (keynum < 0)
			keynum = 0;

		if ((!CSSHost.keys_present) || (CSSHost.titlekeys <= 0))
			return -1;

		// decrypt
		CSSDescramble (lb, CSSHost.titlekey[keynum].key);
	}

	return 0;
}


/**
 * TODO: untested!
 * Encrypt DVD blocks
 * lb_number: LBA of the first block
 * blockcount: number of blocks
 * buffer: pointer to the blocks
 * returns 0 if everything went well, 
 *   -1 on error (failed to encrypt)
 **/

int CSSDriveEncrypt (u_int32_t lb_number, u_int blockcount, u_int8_t *buffer) 
{
	int i, keynum, padding;
	u_int8_t *lb;
	u_int32_t lba;
	static u_int8_t packstartcode[] = {0x00, 0x00, 0x01, 0xBA};

	for (i = 0; i < blockcount; i++) {
		lba = lb_number + i;
		lb = &buffer[i * 2048];

		// check for pack start code at block start
		if (memcmp(lb, packstartcode, 4)) continue;
		padding = lb[13] & 0x07;

		// check for start code (0x000001) at packet start
		if (memcmp(&lb[padding + 14], packstartcode, 3)) continue;

		// video, audio or SPU packet?
		if ((lb[padding + 17] != 0xE0) && 
			(lb[padding + 17] != 0xBD) && 
			((lb[padding + 17] < 0xC0) || (lb[padding + 17] > 0xDF)))
			continue; 

		// PES already encrypted?
		if (lb[padding + 20] & 0x30)
			continue;

		// which title key to use?
		keynum = CSSDrive.titlekeys - 1;
		while ((keynum >= 0) && (CSSDrive.titlekey[keynum].lba > lba))
			keynum--;

		if (keynum < 0)
			continue;

		if ((!CSSDrive.keys_present) || (CSSDrive.titlekeys <= 0))
			return -1;

		// TODO: is that right? is CSS fully symetric?
		CSSDescramble (lb, CSSDrive.titlekey[keynum].key);

		// mark PES encrypted
		lb[padding + 20] = (lb[padding + 20] & 0xCF) | 0x10;

	}

	return 0;
}



/**
 * CSS initialisation
 *
 * converts plain player keys into format used by css-auth package

int CSSInitPlayerKeys(void) 
{
	int i;
	for (i = 0; i < 408; i++) {
		playkeys[i] = malloc(sizeof(playerkey_t));
		if (!playkeys[i])
			return 1;

		playkeys[i]->offset = (i + 1) * LEN_KEY;
		memcpy(playkeys[i]->key, &playerkeys[playkeys[i]->offset], LEN_KEY);
	}
	playkeys[408] = NULL;

	return 0;
}
*/


/**
 * resets the host parameters
 **/

void CSSHostReset(void) 
{
	CSSHost.keys_present = 0;
	CSSHost.titlekeys = 0;
	CSSHost.authOK = 0;
//	CSSInitPlayerKeys();
}


/**
 * resets the drive parameters
 **/

void CSSDriveReset(void) 
{
	int i;

	for (i = 0; i < 4; i++)
		CSSDrive.agid_busy[i] = 0;

	CSSDrive.keys_present = 0;
	CSSDrive.titlekeys = 0;
	CSSDrive.authOK = 0;
}

