/*****
 *
 * This file is part of libcss.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****/

#ifndef __CSS_DVD_H__
#define __CSS_DVD_H__

#include <sys/types.h>

#define CSS_DVD_VERSION 20000214


/***************************************/
/* Load and save key set from/to disc  */
/***************************************/

// reads disc and title keys from a file with the extension .CSS
// returns 0 on success, -1 on error
int CSSDriveReadKeys(char *Filename);

int CSSHostWriteKeys(char *Filename);



/***************************************/
/* drive authorisation: AGID           */
/***************************************/

// Invalidates previously released AGID
// returns 0 on success, 1 on error
int CSSDriveInvalidateAGID(int agid);


// Initializes the CSS process with the drive
// returns positive AGID on success, <0 on error
int CSSDriveRequestAGID(void);

// Reset auth process for host
// returns 0 on success
int CSSHostInitAuth(void);



/***************************************/
/* drive authorisation: c&r            */
/***************************************/

// Host part:
// retreive 10 byte challenge from host for drive
// returns 0 on success, 1 on error
int CSSDriveAuthChallenge(u_int8_t *data);

// Drive part:
// Post host challenge (10 bytes) into drive
// and retreive drive response (5 bytes) for host
// returns 0 on success, 1 on error
int CSSDriveAuth(u_int8_t *data);

// Host part:
// Verify 5 bytes response from drive
// returns 0 on success, 1 on failure
int CSSDriveAuthResponse(u_int8_t *data);



/***************************************/
/* host authorisation: c&r             */
/***************************************/

// Drive part:
// Retreive drive challenge (10 bytes) 
// returns 0 on success, 1 on error
int CSSHostAuthChallenge(u_int8_t *data);

// Host part:
// Post drive challenge (10 bytes) into host
// and retreive host response (5 bytes) for drive
// returns 0 on success, 1 on error
int CSSHostAuth(u_int8_t *data);

// Drive part:
// Post host response (5 bytes) into drive
// returns 0 on success, 1 on error
int CSSHostAuthResponse(u_int8_t *data);



/***************************************/
/* disc key transfer                   */
/***************************************/

// Retreive bus encrypted disc key (2048 byte) from drive
// returns 0 on success, 1 on error
int CSSDriveDiscKey(u_int8_t *data);

// post bus encrypted disc key (2048 byte) into host
// returns 0
int CSSHostDiscKey(u_int8_t *data);



/***************************************/
/* title key transfer                  */
/***************************************/

// Retreive bus encrypted title key (5 byte) from drive
// input parameters:
//   lba: absolute number of logical block containing the title key
// return parameters:
//   data: will be filled with the bus obfuscated title key
//   cpm: 0=material not copyrighted, 1=copyrighted material
//   cp_sec: 1=this sector has specific structure for copyright protection system, 0=it doesn't
//   cgms: 0=copying is permitted, 2=one generation of copies may be made, 3=no copying is allowed
// returns 0 on success, 1 on error
int CSSDriveTitleKey(u_int32_t lba, u_int8_t *data, int *cpm, int *cp_sec, int *cgms);

// post bus encrypted title key (5 byte) into host
// returns 0
int CSSHostTitleKey(u_int32_t lba, u_int8_t *data);



/***************************************/
/* content decoding / encoding         */
/***************************************/

// Decrypts CSS encrypted blocks
// lb_number: LBA of the first block
// blockcount: number of blocks
// buffer: pointer to the blocks
// returns 0 if no encrypted block found, 
//   1 if at least one block got decrypted, 
//   -1 on error (failed to decrypt)
int CSSHostDecrypt (u_int32_t lb_number, u_int blockcount, u_int8_t *buffer);

// TODO: untested!
// Encrypt DVD blocks
// lb_number: LBA of the first block
// blockcount: number of blocks
// buffer: pointer to the blocks
// returns 0 if no block to be encrypted found, 
//   1 if at least one block got encrypted, 
//   -1 on error (failed to encrypt)
int CSSDriveEncrypt (u_int32_t lb_number, u_int blockcount, u_int8_t *buffer);



/***************************************/
/* CSS initialisation                  */
/***************************************/

// resets the host parameters
void CSSHostReset(void);

// resets the drive parameters
void CSSDriveReset(void);

#endif /* DVD_CSS_H */
